// engine.h : main header file for the ENGINE DLL
//

#if !defined(AFX_ENGINE_H__199E2143_C73F_11D2_8045_00105A1E5D53__INCLUDED_)
#define AFX_ENGINE_H__199E2143_C73F_11D2_8045_00105A1E5D53__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#ifndef __AFXWIN_H__
#error include 'stdafx.h' before including this file for PCH
#endif

#include "IOMod.h"
#include "Module.h"
#include "Xpert/Win32Fix.h"
#include "Utils/SSP.h"
#include "Utils/Semaphore.h"
#include "XMLSetup.h"

#undef DECL
#ifdef ENGINE_DLL
#define DECL _declspec(dllexport)
#else
#define DECL _declspec(dllimport)
#endif

extern "C" DECL  bool DebugMode;
extern "C" DECL  bool RebootOnExit;

extern "C" DECL bool GetFileList(CStringArray& List, CString Match);
extern "C" DECL bool GetFileMatch(CString& Name, CString Match);
extern "C" DECL LPCTSTR GetEngineVersion();
extern "C" DECL LPCTSTR GetEnginePath();
extern "C" DECL LPCTSTR GetSetupVersion();

// Base class for Alarm managers
class DECL CAlarmMgr : public CObject
{
public:
   CAlarmMgr();
   virtual ~CAlarmMgr();
   virtual void OnRaiseAlert(int ComPort);
   virtual void OnClearAlert(int ComPort);
   virtual void OnChangeAlarm();
   virtual void OnEnableAlarm();
   virtual void OnDisableAlarm();
   virtual void OnEngineRun();
   virtual void OnEngineStop();
   virtual CString GetStatus();
   virtual bool IsHandlingAlerts(int iPort=0) { return false; }
};

/////////////////////////////////////////////////////////////////////////////
// CSchedule
//
class CSchedule : public CObject
{
   friend class CEngine;
   friend UINT ThreadProc(void * pParam);
public:
   CSchedule(CString Offset, CString Interval, TModule& module);
   CSchedule(CString Offset, CString Interval, CString SampleInterval, 
         CString SampleDuration, TModule& module);
   CSchedule(TTimeSpan Offset, TTimeSpan Interval, TModule& module);
   CSchedule(TTimeSpan Offset, TTimeSpan Interval, TTimeSpan SampleInterval, 
         TTimeSpan SampleDuration, TModule& module);
   virtual ~CSchedule();
   
   TTime NextScheduledSampleTime();
   void NextScheduledStartTime(TTime& tScheduledStart, TTime& tNextSample);
   bool IsScheduleFor(TModule* pMod) {return &m_Module == pMod;}
   
   TTimeSpan GetInterval() {return tsInterval;}
   TTimeSpan GetTime() {return tsOffset;}
   TTimeSpan GetSampleInterval() {return tsSampleInterval;}
   TTimeSpan GetSampleDuration() {return tsSampleDuration;}

private:
   void RunInterval();                       //called by thread proc
   void Exec();                              //starts execution of thread
   BOOL fOverrun;                            //keeps track if an interval overlaps
   TModule& m_Module;                        //ref to module this schedule is for
   TTimeSpan tsInterval;                     //converted interval time
   TTimeSpan tsOffset;                       //converted start time
   TTimeSpan tsSampleInterval;               //converted sample interval time
   TTimeSpan tsSampleDuration;               //converted sample duration time
   HANDLE hStopped;
   HANDLE hEvent;                            // Request an event driven measurement
};


/////////////////////////////////////////////////////////////////////////////
// CScheduleDesc
//
class DECL CScheduleDesc : public CObject
{
public:
   CScheduleDesc();
   CScheduleDesc(const TModule& mod);
   CScheduleDesc& operator=(const CScheduleDesc& sched);
   TTimeSpan tsInterval;
   TTimeSpan tsTime;
   TTimeSpan tsSampleInterval;
   TTimeSpan tsSampleDuration;
   TTimeSpan ComputeDataInterval(); // Compute min interval overlapping timestamp
   TTimeSpan ComputeDataTime();     // Compute min time overlapping timestamp
   // Location of scheduled block relative to finder (-1 upstream, 1 downstream, 0 unknown)
   int iLocation;
}; 


/////////////////////////////////////////////////////////////////////////////
// CTag
//
class DECL CTag : public CObject
{
public:
   CTag(CString NewName);
   CTag();
   virtual ~CTag();
   void SetName(CString NewName);
   bool CheckName(CString strName);

// Tag Functions for Telemetry
public:
   virtual int GetNumValues() = 0;
   virtual int GetAlarm() = 0;
   virtual void SetAlarm(int AlarmState) = 0;
   virtual bool GetTag(int ValueNumber, int& DataType, TValue& Data, CSensorData::QualityType& Quality) = 0;
   virtual bool SetTag(int ValueNumber, int DataType, TValue Data, CSensorData::QualityType Quality)  = 0;
   virtual bool StartTag() = 0;
   virtual bool StopTag() = 0;
   virtual bool EvalTag() = 0;
   virtual bool IsCurDataTag() = 0;
   virtual bool IsViewableTag() = 0;
   virtual bool IsEditableByUserGroup(int iGroup) {return iGroup == SETUP_MODE || iGroup >= CUSTOM_SETUP_GROUP1;}

protected:
   CString Name;
};

//
// CThreadState can be used to keep track of state information on a thread by thread basis
// for each thread which uses the Engine.
//
// So far we are handling a subtle case of detecting whether a thread should abort.
// Some threads need to abort execution when the engine Stops, while other threads
// only need to worry about aborting if the entire application is being terminated.
// In particular, a function which can be called by a scheduled event should abort
// when the engine is no longer running, whereas if the same function is called by
// say the GUI or some other process it shouldn't abort unless the entire application
// is being shutdown.
//
class DECL CThreadState
{
public:
   CThreadState();
   ~CThreadState();
   bool GetScheduled();
   void SetScheduled(bool On);
   HANDLE GetAbortEvent();
   void SetAbortEvent(HANDLE hEvent);
private:
   bool bIsScheduled;
   HANDLE hAbortEvent;
};

/////////////////////////////////////////////////////////////////////////////
// CEngine, controls schedules and module list
//
// There is a single Engine object in the entire application. The Engine is 
// responsible for keeping the list of modules, the list of IO devices, and 
// the list of schedule objects. The Engine also handles the starting and 
// stopping of the system.
class DECL CEngine
{
   ////////////////////////////////
   // Interface
   ////////////////////////////////
public:

   // Constructor and destructor.
   CEngine();
   virtual ~CEngine();

   // List of i2c devices.
   CIOModList IOModList;

   // List of schedule objects. Each creates a thread.
   TObArray<CSchedule> ScheduleList;

   // This is the global object list exported from the engine to the world.
   TObArray<TModule> ModuleList;

   // This is the global object list exported from the engine to the world.
   TObArray<CAlarmMgr> AlarmMgrList;

   // This is a global tag list exported from the engine to the world.
   TMapStringToRefUpperCase<CTag> TagList;
   
   // Signaled when engine is running. Used by threads to identify engine start.
   HANDLE hStartEvent;

   // Signaled when engine is stopped. Used by threads to identify engine stop.
   HANDLE hStopEvent;

   // Signaled when the engine object is being destroyed.
   HANDLE hShutdownEvent;

   // Add a schedule entry to the schedule list.
   BOOL SetSchedule(CString Offset, CString Interval, TModule& module);
   BOOL SetSchedule(CString Offset, CString Interval, CString SampleInterval, 
         CString SampleDuration, TModule& module);
   BOOL SetSchedule(TTimeSpan Offset, TTimeSpan Interval, TModule& module);
   BOOL SetSchedule(TTimeSpan Offset, TTimeSpan Interval, TTimeSpan SampleInterval, 
         TTimeSpan SampleDuration, TModule& module);
   
   // Forces a schedule to execute now (event driven measurements)
   bool ForceSchedule(const TModule& Module);
   
   // Start the engine running.
   void Run(bool bShowErrors=true);
   
   // Since GUI only tracks inputs while user is editing, this routine is used 
   // to build the output lists for each module at runtime.
   void BuildOutputs();
   
   // Stop the engine from running.
   void Stop();

   // Used to indicate the entire system is shutting down
   void Shutdown();

   // Returns true if the engine is running.
   BOOL IsRunning() {return fRunning;};

   // Returns the clock time of the last engine start, regardless of whether 
   // recording is currently on. If recording has never been on, value is 
   // undefined. 
   TTime GetLastRunTime() { return m_tRecStart; };
   
   // Returns the total time span of the current recording session, or 0 if 
   // not running.
   TTimeSpan GetRunningTime(); 
   
   // Station data attributes and methods.
   CString StationName;
   CString MasterID;
   void LoadStationData(void);
   void SaveStationData(void);

   // Call before accessing any GUI elements shared by threads. "Locks" GUI for 
   // exclusive access (blocks until GUI is accessible exclusively). Returns 
   // false if GUI disabled or times out (see DisableGUI, below).
   bool LockGUI(DWORD dwWait = 0);
   
   // Call when finished accessing any GUI elements shared by threads. Releases 
   // "lock" on GUI.
   void UnlockGUI();
   
   // Call to signal GUI is running.
   void EnableGUI();
   
   // Call to signal GUI is stopped.
   void DisableGUI();

   // Used to get access to the tag list or any other dynamic list that may 
   // change during operation
   void LockTags();

   // Call when done accessing the tag list
   void UnLockTags();

   // Used to get access to the alarm mgr list
   void LockAlarms();

   // Call when done accessing the tag list
   void UnLockAlarms();

   // This lock allows concurrent reads but ensures exclusive write access. 
   TRWLock m_rwSetupLock;
   bool LockSetup(DWORD dwWait = 0, bool bReadOnly = true);
   void UnlockSetup(bool bReadOnly = true);
   
   // Call to save setup unconditionally to whatever name is currently set 
   // (defaults to station name, changes as files are opened and/or saved).
   // True is returned if save was successful.
   bool AutoSaveSetup(DWORD dwWait = 0);

   // Posts a message to GUI telling it to save the setup. The setup has not 
   // necessarily been saved yet on return from this method. 
   void SignalSaveSetup();

// Tag and SSP Support functions
   bool BuildTagList(TSSPData& Data, TProtoOpcode Opcode);
   bool WriteValue(TSSPData& Data, int DataType, TValue& TagData);
   bool ReadValue(TSSPData& Data, int& DataType, TValue& TagData);

// Alarm Mangement functions
   bool InAlarm();
   bool InAlert(int ComPort=0);
   void RaiseAlert(int ComPort=0);
   void ClearAlert(int ComPort=0);
   void ChangeAlarm(bool Alarming);
   void ClearAlarm();
   void EnableAlarm();
   void DisableAlarm();
   bool AlarmsEnabled();
   CString GetAlarmStatus();

   // Start and stop Satlink.
   HINSTANCE m_hSatlink;
   bool StartSatlink();
   bool StopSatlink();

   // Returns true if hardware platform is Xpert, false if Xlite.
   bool IsXpert();

   // Measure battery voltage and internal temperature.
   bool GetBattAndTemp(double& dBattVolt, double& dTempC, bool bTempOnly = false);

   // Read and Write the parts of the setup used by Engine
   bool ReadSetup(CXMLSetup& XMLSetup);
   bool WriteSetup(CXMLSetup& XMLSetup);

   ////////////////////////////////
   // Implementation
   ////////////////////////////////
protected:
   // True if engine is running.
   BOOL fRunning;
   TTime m_tRecStart;
   INT64 m_iRecStart;
   DWORD AlertMask;
   bool bAlarmsEnabled;
   DWORD AlarmCount;
   DWORD LastAlarmCount;
   bool LastAlarmStatus;

private:
   friend CThreadState;
   // Attributes used to control access to GUI.
   bool GuiEnabled;
   TSemaphore GuiLock;
   CRITICAL_SECTION TagLock;
   CRITICAL_SECTION AlarmLock;
   CThreadState* DefaultState;
   CObList ThreadStateList;

public:
   // Thead state related functions
   CThreadState& GetThreadState();
   DWORD TlsIndex;
   bool Abort();
   HANDLE GetAbortEvent();
   void (*pFinalStop) ();
};

// The global engine object, used by application and other d/slls
extern "C" DECL CEngine Engine;

class DECL TDigitalModule : public TModule
{
public:
   TDigitalModule(const CString& ModuleName, HINSTANCE hModInst)
      : TModule(ModuleName, hModInst) {};
   ~TDigitalModule()                  {};
   CIODeviceType GetDeviceType() { return DIGITAL; };
   void ClearChannelsInuse()     { Engine.IOModList.ClearChannelInUse(DIGITAL, GetModuleProperty()->AsInteger(), GetChannelProperty()->AsInteger()); };
   void SetChannelsInuse()       { Engine.IOModList.SetChannelInUse(DIGITAL, GetModuleProperty()->AsInteger(), GetChannelProperty()->AsInteger()); };
   bool GetIOAddrAsString(CString& string) { return Engine.IOModList.GetIOAddrAsString(DIGITAL, GetModuleProperty()->AsInteger(), GetChannelProperty()->AsInteger(), string); };
   void AssignDefaultChannel()   { Engine.IOModList.FindChannel(DIGITAL, GetModuleProperty()->AsIntRef(), GetChannelProperty()->AsIntRef()); };
   TProperty* GetModuleProperty()  = 0;
   TProperty* GetChannelProperty() = 0;
};

class DECL TAnalogModule : public TModule
{
public:
   TAnalogModule(const CString& ModuleName, HINSTANCE hModInst)
      : TModule(ModuleName, hModInst) {};
   ~TAnalogModule()                   {};
   CIODeviceType GetDeviceType() { return ANALOG; };
   void ClearChannelsInuse()     { Engine.IOModList.ClearChannelInUse(ANALOG, GetModuleProperty()->AsInteger(), GetChannelProperty()->AsInteger()); };
   void SetChannelsInuse()       { Engine.IOModList.SetChannelInUse(ANALOG, GetModuleProperty()->AsInteger(), GetChannelProperty()->AsInteger()); };
   bool GetIOAddrAsString(CString& string) { return Engine.IOModList.GetIOAddrAsString(ANALOG, GetModuleProperty()->AsInteger(), GetChannelProperty()->AsInteger(), string); };
   void AssignDefaultChannel()   { Engine.IOModList.FindChannel(ANALOG, GetModuleProperty()->AsIntRef(), GetChannelProperty()->AsIntRef()); };
   TProperty* GetModuleProperty()  = 0;
   TProperty* GetChannelProperty() = 0;
};

// The global engine dll info structure, used to set resource keys.
extern AFX_EXTENSION_MODULE EngineDLL;

#endif // !defined(AFX_ENGINE_H__199E2143_C73F_11D2_8045_00105A1E5D53__INCLUDED_)
